{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "---\n",
        "title: \"🗓️ Week 4 - Classification & Metrics & Resampling Methods\"\n",
        "subtitle: \"Theme: Supervised Learning\"\n",
        "author: \"Dr. Ghita Berrada\"\n",
        "date: \"14 February 2025\"\n",
        "format: \n",
        "  html:\n",
        "    from: markdown+emoji\n",
        "    page-layout: full\n",
        "    toc: true\n",
        "    toc-depth: 2\n",
        "self-contained: true\n",
        "---\n",
        "\n",
        "Welcome to Week 4! This week, we go back to classification. But, before we start, let's do a bit of set up\n",
        "\n",
        "⚙️ Set up\n",
        "\n",
        "**Loading Python libraries**"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 1,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "\n",
              "            <div id=\"ktIpJc\"></div>\n",
              "            <script type=\"text/javascript\" data-lets-plot-script=\"library\">\n",
              "                if(!window.letsPlotCallQueue) {\n",
              "                    window.letsPlotCallQueue = [];\n",
              "                }; \n",
              "                window.letsPlotCall = function(f) {\n",
              "                    window.letsPlotCallQueue.push(f);\n",
              "                };\n",
              "                (function() {\n",
              "                    var script = document.createElement(\"script\");\n",
              "                    script.type = \"text/javascript\";\n",
              "                    script.src = \"https://cdn.jsdelivr.net/gh/JetBrains/lets-plot@v4.5.2/js-package/distr/lets-plot.min.js\";\n",
              "                    script.onload = function() {\n",
              "                        window.letsPlotCall = function(f) {f();};\n",
              "                        window.letsPlotCallQueue.forEach(function(f) {f();});\n",
              "                        window.letsPlotCallQueue = [];\n",
              "                        \n",
              "                    };\n",
              "                    script.onerror = function(event) {\n",
              "                        window.letsPlotCall = function(f) {};    // noop\n",
              "                        window.letsPlotCallQueue = [];\n",
              "                        var div = document.createElement(\"div\");\n",
              "                        div.style.color = 'darkred';\n",
              "                        div.textContent = 'Error loading Lets-Plot JS';\n",
              "                        document.getElementById(\"ktIpJc\").appendChild(div);\n",
              "                    };\n",
              "                    var e = document.getElementById(\"ktIpJc\");\n",
              "                    e.appendChild(script);\n",
              "                })()\n",
              "            </script>\n",
              "            "
            ]
          },
          "metadata": {},
          "output_type": "display_data"
        }
      ],
      "source": [
        "import numpy as np\n",
        "import pandas as pd\n",
        "from sklearn.metrics import confusion_matrix, precision_score, recall_score, accuracy_score,f1_score,balanced_accuracy_score\n",
        "from sklearn.neighbors import KNeighborsClassifier\n",
        "import statsmodels.api as sm\n",
        "import sweetviz as sv\n",
        "import missingno as msno\n",
        "from sklearn.preprocessing import StandardScaler\n",
        "from sklearn.preprocessing import LabelEncoder\n",
        "import matplotlib.pyplot as plt\n",
        "import seaborn as sns\n",
        "from lets_plot import *\n",
        "LetsPlot.setup_html()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "# Part 1: A tennis match outcome prediction problem\n",
        "\n",
        "## Introducing the dataset\n",
        "\n",
        "In this part, we'll be looking at a dataset about tennis^[credit to Vignesh for the idea thanks to his mention of GOAT on Slack😉]. The data contains match stats and player info for WTA (Women's Tennis Association) and ATP (Association of Tennis Professionals) between 1949 and 2021. The data is available on [Kaggle](https://www.kaggle.com/datasets/taylorbrownlow/atpwta-tennis-data?resource=download) and was originally sourced from [Jeff Sackmann's GitHub repository](https://github.com/JeffSackmann/tennis_atp) (which maintains the dataset and updates it on a regular basis)."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 5,
      "metadata": {},
      "outputs": [],
      "source": [
        "df = pd.read_csv('../data/KaggleMatches.csv',low_memory=False)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "For simplicity, we'll filter the original dataset so as to only retain ATP data. "
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 6,
      "metadata": {},
      "outputs": [],
      "source": [
        "# filtering to only retain ATP data\n",
        "df=df.query(\"league=='atp'\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "We'll also filter the data to only retain data from 1969 onwards. \n",
        "\n",
        "Any ideas as to why we do that? \n",
        "\n",
        "And how do we do that?"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Let's have a look at our data:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "report = sv.analyze(filtered_df)\n",
        "report.show_notebook() "
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "What features do you think we can remove straight away? Why? Which should we transform for modeling purposes?\n",
        "\n",
        "\n",
        "\n",
        "## 1.2 Before we continue: a recap on logistic regression\n",
        "\n",
        "Consider a **binary** response:\n",
        "\n",
        "$$\n",
        "Y = \\begin{cases}\n",
        "0 \\\\\n",
        "1\n",
        "\\end{cases}\n",
        "$$\n",
        "\n",
        "We model the probability that $Y = 1$ using the **logistic function** (aka. sigmoid curve):\n",
        "\n",
        "$$\n",
        "Pr(Y = 1|X) = p(X) = \\frac{e^{\\beta_0 + \\beta_1X}}{1 + e^{\\beta_0 + \\beta_1 X}}\n",
        "$$\n",
        "\n",
        "This function is mathematically guaranteed to be bounded by the interval $[0, 1]$.\n",
        "\n",
        "### 1.1.1 Insights into the logistic regression function\n",
        "\n",
        "Let's see how this mathematical function looks like:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "def plot_logistic_function(X_range, beta0_values, beta1_values, color=None):\n",
        "    # Create a dataframe similar to expand.grid() in R\n",
        "    df = pd.DataFrame(\n",
        "        [(X, beta0, beta1) for X in X_range for beta0 in beta0_values for beta1 in beta1_values],\n",
        "        columns=['X', 'beta0', 'beta1']\n",
        "    )\n",
        "\n",
        "    # Compute logistic function\n",
        "    df['Y'] = np.exp(df['beta0'] + df['beta1'] * df['X']) / (1 + np.exp(df['beta0'] + df['beta1'] * df['X']))\n",
        "\n",
        "    # Convert to categorical for ordered factors\n",
        "    df['beta0'] = df['beta0'].astype(str)\n",
        "    df['beta1'] = df['beta1'].astype(str)\n",
        "\n",
        "    # Choose color mapping\n",
        "    if color is None:\n",
        "        p = ggplot(df, aes(x='X', y='Y'))\n",
        "    elif color == \"beta0\":\n",
        "        p = ggplot(df, aes(x='X', y='Y', color='beta0', linetype='beta0'))\n",
        "    else:\n",
        "        p = ggplot(df, aes(x='X', y='Y', color='beta1', linetype='beta1'))\n",
        "\n",
        "    # Add lines and customize appearance\n",
        "    p += geom_line(size=1.4) + \\\n",
        "        ggsize(800, 400) + \\\n",
        "        labs(x=\"X\", y=\"Y\") + \\\n",
        "        theme(panel_grid_minor='blank') + \\\n",
        "        theme(axis_title=element_text(size=24), \n",
        "              axis_text=element_text(size=18), \n",
        "              legend_text=element_text(size=18),\n",
        "              legend_position=\"right\")\n",
        "\n",
        "    return p"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "X_range = np.linspace(-400, 400, 100)\n",
        "beta0_values = [-3, 0, 1]\n",
        "beta1_values = [1, 1, 1]\n",
        "p=plot_logistic_function(X_range, beta0_values, beta1_values, color=\"beta1\")\n",
        "p+=ggtitle(\"If I keep beta1 constant and equal to 1 and vary beta0\")\n",
        "p"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "X_range = np.linspace(-4, 4, 100)\n",
        "beta0_values = [0, 0, 0, 0]\n",
        "beta1_values = [-1, 0, 1, 2]\n",
        "\n",
        "p=plot_logistic_function(X_range, beta0_values, beta1_values, color=\"beta1\")\n",
        "p+=ggtitle(\"If I keep beta0 constant and equal to 0 and vary beta1\")\n",
        "p"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## 1.2 Going back to our match prediction problem\n",
        "\n",
        "I, on the sly, define my tennis match lose/win prediction problem as a binary class prediction problem: the winner is the player with the most ranking points (regardless of his name!). My outcome variable is the aptly named `outcome`.\n",
        "\n",
        "I have temporal data so I can't do my training/test split as usual. I'll keep data prior to 2020 as training data and data posterior to 2020 as test data"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "train_df = filtered_df[filtered_df['tourney_date'] < '2020-01-01']\n",
        "test_df = filtered_df[filtered_df['tourney_date'] >= '2020-01-01']"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Let's try and fit a logistic regression model on the training data:\n",
        "\n",
        "```{r}\n",
        "\n",
        "```\n",
        "\n",
        "\n",
        "```text\n",
        "Optimization terminated successfully.\n",
        "         Current function value: 0.607134\n",
        "         Iterations 5\n",
        "                           Logit Regression Results                           \n",
        "==============================================================================\n",
        "Dep. Variable:                outcome   No. Observations:                 6671\n",
        "Model:                          Logit   Df Residuals:                     6644\n",
        "Method:                           MLE   Df Model:                           26\n",
        "Date:                Fri, 14 Feb 2025   Pseudo R-squ.:                 0.05360\n",
        "Time:                        12:02:41   Log-Likelihood:                -4050.2\n",
        "converged:                       True   LL-Null:                       -4279.6\n",
        "Covariance Type:            nonrobust   LLR p-value:                 1.131e-80\n",
        "===============================================================================\n",
        "                  coef    std err          z      P>|z|      [0.025      0.975]\n",
        "-------------------------------------------------------------------------------\n",
        "const           0.5121      0.127      4.029      0.000       0.263       0.761\n",
        "surface         0.0555      0.031      1.778      0.075      -0.006       0.117\n",
        "winner_hand     0.0271      0.077      0.352      0.725      -0.124       0.178\n",
        "winner_ht       0.0476      0.026      1.819      0.069      -0.004       0.099\n",
        "winner_age     -0.0481      0.027     -1.812      0.070      -0.100       0.004\n",
        "loser_hand     -0.0489      0.077     -0.637      0.524      -0.199       0.102\n",
        "loser_ht       -0.0157      0.027     -0.588      0.557      -0.068       0.037\n",
        "loser_age       0.1104      0.028      3.974      0.000       0.056       0.165\n",
        "best_of         0.0378      0.039      0.976      0.329      -0.038       0.114\n",
        "w_ace           0.1015      0.035      2.922      0.003       0.033       0.170\n",
        "...\n",
        "l_SvGms        -0.4567      0.216     -2.115      0.034      -0.880      -0.034\n",
        "l_bpSaved       0.0705      0.167      0.423      0.672      -0.256       0.397\n",
        "l_bpFaced      -0.0584      0.196     -0.298      0.766      -0.443       0.326\n",
        "===============================================================================\n",
        "```\n",
        "\n",
        "What does this tell me?\n",
        "\n",
        "- If variables were 0, the log-odds of a win increases by the intercept i.e 0.5121.\n",
        "\n",
        "The probability of a win increases by:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "log_odds = 0.5121\n",
        "probability = 1 / (1 + np.exp(-log_odds))\n",
        "print(probability)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "- If all other predictors are held constant, wach additional ace hit by the winner increases the odds of winning bu 0.1015 and the probability of winning increases by:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "w_ace = 1\n",
        "w_ace_coef = 0.1015\n",
        "new_log_odds = 0.5121 + (w_ace_coef * w_ace)\n",
        "\n",
        "probability_w_ace_1 = 1 / (1 + np.exp(-new_log_odds))\n",
        "print(probability_w_ace_1)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "How should I go about evaluating the performance of my model?"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "X_test = test_df.drop(columns=[\"outcome\",\"tourney_name\",\"winner_name\",\"loser_name\",\"tourney_level\",'round','winner_rank','winner_rank_points','loser_rank','loser_rank_points','tourney_date'], axis=1).astype(float)\n",
        "X_test = sm.add_constant(X_test)\n",
        "y_test = test_df[\"outcome\"].astype(int)\n",
        "\n",
        "y_pred = logit_model.predict(X_test)\n",
        "\n",
        "# Convert probabilities to binary outcome (0 or 1) using threshold of 0.5\n",
        "y_pred_binary = (y_pred > 0.5).astype(int)\n",
        "\n",
        "# Confusion Matrix\n",
        "conf_matrix = confusion_matrix(y_test, y_pred_binary)\n",
        "print(\"Confusion Matrix:\")\n",
        "print(conf_matrix)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Let's visualise the same thing a bit better:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "conf_matrix = confusion_matrix(y_test, y_pred_binary)\n",
        "\n",
        "# Create a heatmap\n",
        "plt.figure(figsize=(6, 5))  # Size of the plot\n",
        "sns.heatmap(conf_matrix, annot=True, fmt='g', cmap='Blues', cbar=False, \n",
        "            xticklabels=[\"Predicted Loss\", \"Predicted Win\"], \n",
        "            yticklabels=[\"Actual Loss\", \"Actual Win\"])\n",
        "plt.title(\"Confusion Matrix Heatmap\")\n",
        "plt.xlabel(\"Predicted Label\")\n",
        "plt.ylabel(\"True Label\")\n",
        "plt.show()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Can you do the same on the training set? What do you notice?\n",
        "\n",
        "\n",
        "What I have here is called a **confusion matrix**: it represents counts from predicted and actual values.\n",
        "\n",
        "Perhaps, a bit of vocabulary to know in the context of binary classification:\n",
        "\n",
        "-   the two classes in your classification are called positive class (i.e the class you're interested in, here match wins), and negative class\n",
        "\n",
        "-   knowing that, you define true positives (TP) as the instances in your dataset that are of the positive class and are correctly classified by the model as belonging to the positive class, true negatives (TN) are the instances that are of the negative class and are correctly predicted by the model as belonging to the negative class, false positives (FP) are instances of the negative class that are misclassified by the model as instances of the positive class and false negatives (FN) are instances of the positive class that are falsely predicted by the model as belonging to the negative class\n",
        "\n",
        "-   whether the FN or FP count is more important depends on your application\n",
        "\n",
        "Now, let's have a look at a few more metrics:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Precision\n",
        "precision = precision_score(y_test, y_pred_binary)\n",
        "print(\"Precision:\", precision)\n",
        "\n",
        "# Recall\n",
        "recall = recall_score(y_test, y_pred_binary)\n",
        "print(\"Recall:\", recall)\n",
        "\n",
        "f1_score_test = f1_score(y_test, y_pred_binary)\n",
        "print(\"F1 score:\", f1_score_test)\n",
        "\n",
        "# Accuracy\n",
        "accuracy = accuracy_score(y_test, y_pred_binary)\n",
        "print(\"Accuracy:\", accuracy)\n",
        "\n",
        "# Class Imbalance: Proportion of each class in the target variable\n",
        "class_distribution = np.bincount(y_test) / len(y_test)\n",
        "print(\"Class Imbalance (Proportions):\")\n",
        "print(f\"Class 0 (Loss): {class_distribution[0]:.2f}, Class 1 (Win): {class_distribution[1]:.2f}\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "-   Accuracy\n",
        "\n",
        "    It's the proportion of correct predictions i.e\n",
        "\n",
        "    $$\n",
        "    \\begin{equation}\n",
        "    accuracy = \\frac{TP+TN}{P+N}\\\n",
        "    \\end{equation}\n",
        "    $$\n",
        "\n",
        "    where $P$ is the total number of positive instances in the dataset and $N$ the total number of negative instances in the dataset.\n",
        "\n",
        "    It's only a good metric when the classes within your dataset are balanced. Can you guess what the problem might be in case of imbalance?\n",
        "\n",
        "    How do you check if classes are balanced?\n",
        "\n",
        "What about our data? The win class is over-represented!\n",
        "\n",
        "So accuracy is **not** an appropriate metric in this case.\n",
        "\n",
        "What metrics can we use in case of imbalance?\n",
        "\n",
        "-   Precision, recall/sensitivity and F1-score\n",
        "\n",
        "    Precision is the proportion of true positives among all the predicted positive cases.\n",
        "\n",
        "    $$\n",
        "    \\begin{align}\n",
        "    precision &= \\frac{TP}{TP+FP}\\\\\n",
        "              &= \\frac{TP}{\\textrm{all predicted positives}}\n",
        "    \\end{align}\n",
        "    $$\n",
        "\n",
        "    It is a good metric when the cost of false positives is high, e.g spam detection (the consequences of misclassifying a legitimate email as spam can be significant as it may result in important communications being filtered out and so the cost of a false positive is higher than that of a false negative in this case).\n",
        "\n",
        "    Recall (or sensitivity) is the proportion of true positives among all actual positive cases.\n",
        "\n",
        "    $$\n",
        "    \\begin{align}\n",
        "    recall &= \\frac{TP}{TP+FN}\\\\\n",
        "              &= \\frac{TP}{\\textrm{all actual positives}}\n",
        "    \\end{align}\n",
        "    $$\n",
        "\n",
        "    High recall is critical in applications where the cost of false negatives is high, e.g medical applications. If you think back to this week's lab, a false negative meant a patient that had diabetes being misdiagnosed as being healthy resulting in delayed diagnosis and delayed appropriate treatment (and likely deterioration of the patient's condition in the meanwhile).\n",
        "\n",
        "    Most applications, though, will need a balance between precision and recall. That's where the F-score (especially in the form of the balanced F1-score) comes in: it's the harmonic mean of precision and recall.\n",
        "\n",
        "    $$\n",
        "    \\textrm{F1-score}=2\\frac{precision * recall}{precision + recall}\n",
        "    $$\n",
        "\n",
        "    A more general score $F_\\beta$ exists if you want to emphasize precision over recall or vice versa:\n",
        "\n",
        "    $$\n",
        "    \\textrm{F}_{\\beta}\\textrm{-score} = (1+\\beta^2)\\frac{precision * recall}{(\\beta^2.precision)+ recall}\n",
        "    $$\n",
        "\n",
        "-   Specificity\n",
        "\n",
        "    The specificity is the proportion of negatives that are correctly identified as negatives.\n",
        "\n",
        "    $$\n",
        "    specificity = \\frac{TN}{TN+FP}\n",
        "    $$\n",
        "\n",
        "    It is important in contexts where the classification of negatives is of high priority.\n",
        "\n",
        "    It's particularly useful in the medical context. Consider the example of a medical test for diagnosing a disease; e.g the HbA1C blood test used to diagnose diabetes. Specificity refers to the test's ability to correctly reject healthy patients without the condition (i.e without diabetes). Supposing HbA1C had high specificity, a positive result in such test with high specificity would be useful for \"ruling in\" disease, i.e determining that the patient does in fact have diabetes, since the test rarely gives positive results in healthy patients.\n",
        "\n",
        "-   Balanced accuracy\n",
        "\n",
        "    But specificity is also used in another context. It is one of the components used to define balanced accuracy, a metric designed to overcome the problem that accuracy has when it comes to imbalanced data.\n",
        "\n",
        "    $$\n",
        "    \\begin{equation}\n",
        "    \\textrm{balanced}\\_\\textrm{accuracy}= \\frac{sensitivity+specificity}{2}\n",
        "    \\end{equation}\n",
        "    $$\n",
        "\n",
        "The closer the balanced accuracy is to 1, the better the model is able to correctly classify observations (for both classes).\n",
        "\n",
        "**A recap of the metrics**\n",
        "\n",
        "| **Metric** | **Description** |\n",
        "|---------------------------------|--------------------------------------|\n",
        "| Accuracy | Proportion of correct predictions |\n",
        "| Sensitivity (=recall) | The proportion of positive results out of the number of samples that were actually positive. |\n",
        "| Specificity | The proportion of negatives that are correctly identified as negatives. |\n",
        "| Precision | The proportion of true positive results out of the number of samples that were predicted as positive. |\n",
        "| Recall (=sensitivity) | The proportion of true positive results out of the number of samples that were actually positive. |\n",
        "| F-measure | The harmonic mean of precision and recall. |\n",
        "| Balanced accuracy | Mean of sensitivity and specificity |\n",
        "\n",
        "# Part 2: Classifying differently: K-NN\n",
        "\n",
        "## 2.1 The principles\n",
        "\n",
        "**Basic idea (images and example from [here](https://www.analyticsvidhya.com/blog/2018/03/introduction-k-neighbours-algorithm-clustering/)):**\n",
        "\n",
        "Look at this example. Say you want to predict the class of the blue star (BS) and see whether it's a red circle (RC) or a green square (GS).\n",
        "\n",
        "![](images/paste-4.png)\n",
        "\n",
        "$K$-NN will take \"votes\" from the nearest neighbours (nearest defined based on pre-specified distance usually Euclidian distance) and will only look at $K$ neighbours (hence the $K$ in $K$-NN).\n",
        "\n",
        "In this case, if we take $K=3$, we see that the blue star is surrounded by 3 red circles and no green squares, so the star would be classified as a red circle.\n",
        "\n",
        "![](images/paste-5.png)\n",
        "\n",
        "**K-NN: the steps**\n",
        "\n",
        "1.  Load the data\n",
        "\n",
        "2.  Initialise the value of \\$K\\$\n",
        "\n",
        "3.  To get the predicted class, iterate over the test data points\n",
        "\n",
        "    -   Calculate the distance between test data point and each row of the training set. By default, use the Euclidian distance. Other distance metrics that could be used are the Manhattan distance, Minkowski distance, Chebyshev, cosine, etc. If there are categorical variables, the Hamming distance can be used.\n",
        "\n",
        "    -   Sort the calculated distances in ascending order based on distance values\n",
        "\n",
        "    -   Get top $K$ rows from the sorted array of distances\n",
        "\n",
        "    -   Get the most frequent class from these rows\n",
        "\n",
        "    -   Return the most frequent class as the predicted class for the current test data point\n",
        "\n",
        "## 2.2 Applying it to our problem of predicting tennis match wins\n",
        "\n",
        "How about we try $K$-NN to the same prediction problem as before: predicting tennis match wins"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "knn = KNeighborsClassifier(n_neighbors=3)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Let's evaluate the model now, first on the training set then the test set:"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "knn.fit(X, y)\n",
        "\n",
        "y_pred_train_knn = knn.predict(X)\n",
        "# Predict on test data\n",
        "y_pred_knn = knn.predict(X_test)\n",
        "\n",
        "# Display results\n",
        "print(\"Predicted Rounds (Encoded):\", y_pred_knn)\n",
        "print(\"Actual Rounds (Encoded):\", y_test.values)"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "y_pred_train_binary_knn = (y_pred_train_knn > 0.5).astype(int)\n",
        "\n",
        "# Confusion Matrix\n",
        "conf_matrix_train_knn = confusion_matrix(y, y_pred_train_binary_knn)\n",
        "print(\"Confusion Matrix (training set):\")\n",
        "print(conf_matrix_train_knn)\n",
        "\n",
        "# Precision\n",
        "precision_train_knn = precision_score(y, y_pred_train_binary_knn)\n",
        "print(\"Precision (training set):\", precision_train_knn)\n",
        "\n",
        "# Recall\n",
        "recall_train_knn = recall_score(y, y_pred_train_binary_knn)\n",
        "print(\"Recall (training set):\", recall_train_knn)\n",
        "\n",
        "# Accuracy\n",
        "f1_score_train_knn = f1_score(y, y_pred_train_binary_knn)\n",
        "print(\"F1 score (training set):\", f1_score_train_knn)\n",
        "\n",
        "balanced_accuracy_train_knn = balanced_accuracy_score(y, y_pred_train_binary_knn)\n",
        "print(\"Balanced accuracy (training set):\", balanced_accuracy_train_knn)\n",
        "\n",
        "# Class Imbalance: Proportion of each class in the target variable\n",
        "class_distribution_train = np.bincount(y) / len(y)\n",
        "print(\"Class Imbalance (Proportions) (training set):\")\n",
        "print(f\"Class 0 (Loss): {class_distribution_train[0]:.2f}, Class 1 (Win): {class_distribution_train[1]:.2f}\")"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {},
      "outputs": [],
      "source": [
        "y_pred_binary_knn = (y_pred_knn > 0.5).astype(int)\n",
        "\n",
        "# Confusion Matrix\n",
        "conf_matrix_knn = confusion_matrix(y_test, y_pred_binary_knn)\n",
        "print(\"Confusion Matrix:\")\n",
        "print(conf_matrix_knn)\n",
        "\n",
        "# Precision\n",
        "precision_knn = precision_score(y_test, y_pred_binary_knn)\n",
        "print(\"Precision:\", precision_knn)\n",
        "\n",
        "# Recall\n",
        "recall_knn = recall_score(y_test, y_pred_binary_knn)\n",
        "print(\"Recall:\", recall_knn)\n",
        "\n",
        "# Accuracy\n",
        "f1_score_test_knn = f1_score(y_test, y_pred_binary_knn)\n",
        "print(\"F1 score:\", f1_score_test_knn)\n",
        "\n",
        "balanced_accuracy_knn = balanced_accuracy_score(y_test, y_pred_binary_knn)\n",
        "print(\"Balanced accuracy:\", balanced_accuracy_knn)\n",
        "\n",
        "# Class Imbalance: Proportion of each class in the target variable\n",
        "class_distribution = np.bincount(y_test) / len(y_test)\n",
        "print(\"Class Imbalance (Proportions):\")\n",
        "print(f\"Class 0 (Loss): {class_distribution[0]:.2f}, Class 1 (Win): {class_distribution[1]:.2f}\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "How does $K$-NN compare to logistic regression?\n",
        "\n",
        "📚 **Notes**: When independent variables in training data are measured in different units, it is important to standardize variables before calculating distance. So, it is usually good practice to normalize the dataset before applying $K$-NN. You can do that with `StandardScaler` for example. You can have a look at [this page](https://www.geeksforgeeks.org/how-to-standardize-data-in-a-pandas-dataframe/) to see ways to standardize your data.\n",
        "\n",
        "Because it relies on distance calculations, $K$-NN is sensitive to outliers\n",
        "\n",
        "🎯 ACTION POINT: How should I go about modifying my $K$-NN model and its evaluation?\n"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "base",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.12.2"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 4
}
